#!/usr/bin/env python3
from collections import ChainMap
from coralillo import Engine
from itacate import Config
from lark.exceptions import GrammarError, ParseError, LexError
from xml.dom import pulldom
from xml.sax._exceptions import SAXParseException
from yuid import yuid
import argparse
import logging
import logging.config
import os
import re
import sys
import time

from cacahuate.errors import MalformedProcess
from cacahuate.grammar import Condition
from cacahuate.indexes import create_indexes
from cacahuate.loop import Loop
from cacahuate.models import bind_models
from cacahuate.xml import NODES, get_text


def main():
    # Load the config
    config = Config(os.path.dirname(os.path.realpath(__file__)))
    config.from_object('cacahuate.settings')
    config.from_envvar('CACAHUATE_SETTINGS', silent=True)

    # Set the timezone
    os.environ['TZ'] = config['TIMEZONE']
    time.tzset()

    # Setup logging
    logging.config.dictConfig(config['LOGGING'])

    # Load the models
    eng = Engine(
        host=config['REDIS_HOST'],
        port=config['REDIS_PORT'],
        db=config['REDIS_DB'],
        id_function=yuid,
    )
    bind_models(eng)

    # Create mongo indexes
    create_indexes(config)

    # start the loop
    loop = Loop(config)
    loop.start()


def rng_path():
    print(os.path.abspath(os.path.join(
        os.path.dirname(__file__),
        'xml/process-spec.rng'
    )))


def _validate_file(filename):
    ids = []
    data_form = ChainMap()
    passed_nodes = []
    variable_re = re.compile(r'^[a-zA-Z_][a-zA-Z0-9_]*$')

    class StreamWrapper:

        def __init__(self, filename):
            self.f = open(filename)
            self.lineno = 0

        def read(self, b):
            self.lineno += 1
            return self.f.readline()

    sw = StreamWrapper(filename)

    def check_id(node):
        id_element = node.getAttribute('id')

        if not id_element:
            raise MalformedProcess(
                '{}:{} All nodes must have an id'.format(filename, sw.lineno)
            )

        if not variable_re.match(id_element):
            raise MalformedProcess(
                '{}:{} Id must be a valid variable name'.format(
                    filename, sw.lineno,
                )
            )

        if id_element not in ids:
            ids.append(id_element)
        else:
            raise MalformedProcess("{}:{} Duplicated id: '{}'".format(
                filename, sw.lineno,
                id_element,
            ))

    doc = pulldom.parse(sw)

    for event, node in doc:
        if event == pulldom.START_ELEMENT and node.tagName == 'block':
            data_form.maps.append(dict())

        if event == pulldom.END_ELEMENT and node.tagName == 'block':
            data_form.maps.pop()

        if event != pulldom.START_ELEMENT:
            continue

        if node.tagName == 'condition':
            doc.expandNode(node)

            try:
                tree = Condition().parse(get_text(node))
            except GrammarError:
                raise MalformedProcess(
                    '{}:{} Grammar error in condition'.format(
                        filename, sw.lineno,
                    )
                )
            except ParseError:
                raise MalformedProcess(
                    '{}:{} Parse error in condition'.format(
                        filename, sw.lineno,
                    )
                )
            except LexError:
                raise MalformedProcess(
                    '{}:{} Lex error in condition'.format(filename, sw.lineno)
                )
            except KeyError as e:
                raise MalformedProcess(
                    '{}:{} variable does not exist: {}'.format(
                        filename, sw.lineno,
                        str(e),
                    )
                )

            # validate variables used in condition
            for tree in tree.find_data('ref'):
                reference_form, field_form = list(map(
                    lambda x: x.children[0][:],
                    tree.children
                ))

                try:
                    data_form[reference_form][field_form]
                except KeyError:
                    raise MalformedProcess(
                        "{}:{} variable used in if is not defined '{}'".format(
                            filename, sw.lineno,
                            reference_form+'.'+field_form,
                        )
                    )

            continue

        if node.tagName not in NODES:
            continue

        # Duplicate ids
        check_id(node)

        if node.tagName == 'if':
            continue

        # Expand and check this node. <if> nodes are not expanded
        doc.expandNode(node)

        # Check auth-filter params
        params = node.getElementsByTagName("param")

        for param in params:
            if param.getAttribute('type') != 'ref':
                continue

            ref_type, ref = get_text(param).split('#')

            if ref_type == 'form':
                reference_form, field_form = ref.split('.')

                try:
                    data_form[reference_form][field_form]
                except KeyError:
                    raise MalformedProcess(
                        "{}:{} Referenced param does not exist '{}'".format(
                            filename, sw.lineno,
                            reference_form+'.'+field_form,
                        )
                    )
            elif ref_type == 'user':
                if ref not in passed_nodes:
                    raise MalformedProcess(
                        '{}:{} Referenced user is never created: {}'.format(
                            filename, sw.lineno,
                            ref,
                        )
                    )

        # Check dependencies
        deps = node.getElementsByTagName("dep")

        for dep in deps:
            form, field = get_text(dep).split('.')

            try:
                data_form[form][field]
            except KeyError:
                raise MalformedProcess(
                    "{}:{} Referenced dependency does not exist '{}'".format(
                        filename, sw.lineno,
                        form + '.' + field,
                    )
                )

        # fill forms for later usage
        forms = node.getElementsByTagName('form')

        for form in forms:
            form_id = form.getAttribute('id')

            if form_id and not variable_re.match(form_id):
                raise MalformedProcess(
                    '{}:{} Form ids must be valid variable names'.format(
                        filename, sw.lineno,
                    )
                )

            form_ref = form.getAttribute('ref')

            if form_ref and not variable_re.match(form_ref):
                raise MalformedProcess(
                    '{}:{} Form refs must be valid variable names'.format(
                        filename, sw.lineno,
                    )
                )

            inputs = form.getElementsByTagName("input")
            array_input = {}

            for inpt in inputs:
                inpt_name = inpt.getAttribute('name')

                if not variable_re.match(inpt_name):
                    raise MalformedProcess(
                        '{}:{} Field names must match [a-zA-Z0-9_]+'.format(
                            filename, sw.lineno,
                        )
                    )

                array_input[inpt_name] = \
                    inpt.getAttribute('default')

            data_form.maps[-1][form_id] = array_input

        # add this node to the list of revised nodes
        has_auth_filter = len(node.getElementsByTagName('auth-filter')) > 0

        if has_auth_filter:
            passed_nodes.append(node.getAttribute('id'))

    print('{} seems correct'.format(filename))


def xml_validate(filenames=None):
    if not filenames:
        parser = argparse.ArgumentParser(description='Validate xmls')

        parser.add_argument('files', metavar='FILE', nargs='+',
                            help='the files to be validated')

        args = parser.parse_args()

        filenames = args.files

    found_errors = False

    for filename in filenames:
        try:
            try:
                _validate_file(filename)
            except SAXParseException:
                raise MalformedProcess('{}:{} Is not valid xml'.format(
                    filename, 0
                ))
            except FileNotFoundError:
                raise MalformedProcess('{} not found'.format(filename))
        except MalformedProcess as e:
            found_errors = True
            print(e)

    if found_errors:
        sys.exit('** Validation errors found **')


if __name__ == '__main__':
    main()



''' Logic on how information is invalidated in cascade. It is used by
validation-type nodes and patch requests '''
from cacahuate.errors import EndOfProcess


def cascade_invalidate(xml, state, invalidated, comment):
    ''' computes a set of fields to be marked as invalid given the
    original `invalidated` set of fields. '''
    # because this could cause a recursive import
    from cacahuate.node import make_node

    # find the first node that is invalid and select it
    set_values = {
        i['ref']: {
            'value': i['value'],
            'value_caption': i['value_caption'],
        }
        for i in invalidated
        if 'value' in i
    }
    invalid_refs = set(
        i['ref']
        for i in invalidated
    )
    xmliter = iter(xml)

    for element in xmliter:
        node = make_node(element, xmliter)
        more_fields = node.get_invalidated_fields(invalid_refs, state)

        invalid_refs.update(more_fields)

    # computes the keys and values to be used in a mongodb update to set the
    # fields as invalid
    updates = dict()

    for key in invalid_refs:
        node, actor, form, input = key.split('.')
        index, ref = form.split(':')

        node_path = 'state.items.{node}'.format(node=node)
        comment_path = node_path + '.comment'
        node_state_path = node_path + '.state'
        actor_path = node_path + '.actors.items.{actor}'.format(actor=actor)
        actor_state_path = actor_path + '.state'
        form_path = actor_path + '.forms.{index}'.format(index=index)
        form_state_path = form_path + '.state'
        input_path = form_path + '.inputs.items.{input}'.format(input=input)
        input_state_path = input_path + '.state'
        input_value_path = input_path + '.value'
        input_caption_path = input_path + '.value_caption'

        # inputs
        input_state = 'valid' if key in set_values else 'invalid'

        updates[input_state_path] = input_state

        if key in set_values:
            updates[input_value_path] = set_values[key]['value']
            updates[input_caption_path] = set_values[key]['value_caption']

        # forms
        if input_state == 'valid' and (
                form_state_path not in updates or
                updates[form_state_path] == 'valid'):
            form_state = 'valid'
        else:
            form_state = 'invalid'

        updates[form_state_path] = form_state

        # actors
        if form_state == 'valid' and (
                actor_state_path not in updates or
                updates[actor_state_path] == 'valid'):
            actor_state = 'valid'
        else:
            actor_state = 'invalid'

        updates[actor_state_path] = actor_state

        # nodes
        if actor_state == 'valid' and (
                node_state_path not in updates or
                updates[node_state_path] == 'valid'):
            node_state = 'valid'
        else:
            node_state = 'invalid'

        updates[node_state_path] = node_state
        updates[comment_path] = comment

    return updates


def track_next_node(xml, state, mongo, config):
    ''' given an xml and the current state, returns the first invalid or
    unfilled node following the xml's ruleset (conditionals) '''
    from cacahuate.node import make_node

    xmliter = iter(xml)
    node = make_node(next(xmliter), xmliter)

    if node.id in state['state']['items']:
        node_state = state['state']['items'][node.id]['state']
        if node_state in ('invalid', 'unfilled'):
            return node

    try:
        while True:
            node = node.next(
                xml,
                state,
                mongo,
                config,
                skip_reverse=True,
            )

            if node.id in state['state']['items']:
                if state['state']['items'][node.id]['state'] == 'valid':
                    continue

            return node
    except StopIteration:
        # End of process
        raise EndOfProcess



from case_conversion import pascalcase
from importlib import import_module
from coralillo.errors import ModelNotFoundError
import os
import sys
from jinja2 import Template, TemplateError

from cacahuate.errors import MisconfiguredProvider
from cacahuate.models import User


def user_import(module_key, class_sufix, import_maper, default_path, enabled):
    ''' import a provider defined by the user '''
    if module_key in import_maper:
        import_path = import_maper[module_key]

        cwd = os.getcwd()

        if cwd not in sys.path:
            sys.path.insert(0, cwd)
    elif module_key in enabled:
        import_path = default_path + '.' + module_key
    else:
        raise MisconfiguredProvider(
            'Provider {} not enabled'.format(module_key)
        )

    cls_name = pascalcase(module_key) + class_sufix

    try:
        mod = import_module(import_path)
        cls = getattr(mod, cls_name)
    except ModuleNotFoundError:
        raise MisconfiguredProvider(
            'Could not import provider module {}'.format(import_path)
        )
    except AttributeError:
        raise MisconfiguredProvider('Provider does not define class {}'.format(
            cls_name,
        ))

    return cls


def clear_username(string):
    ''' because mongo usernames have special requirements '''
    string = string.strip()

    if string.startswith('$'):
        string = string[1:]

    try:
        string = string[:string.index('@')]
    except ValueError:
        pass

    return string.replace('.', '')


def get_or_create(identifier, data):
    identifier = clear_username(identifier)
    data['identifier'] = identifier

    try:
        return User.get_by_or_exception('identifier', identifier)
    except ModelNotFoundError:
        return User(**data).save()


def render_or(template, default, context={}):
    ''' Renders the given template in case it is a valid jinja template or
    returns the default value '''
    try:
        return Template(template).render(**context)
    except TemplateError:
        return default


def compact_values(collected_forms):
    ''' Takes the input from the first form in a process and compacts it into
    the first set of values for _the state_ that will be used in conditionals
    or jinja templates'''
    context = dict()
    for form in collected_forms:
        form_dict = dict()

        for name, input in form['inputs']['items'].items():
            form_dict[name] = input['value_caption']

        context[form['ref']] = form_dict

    return context



from coralillo.errors import ModelNotFoundError
from cacahuate.http.wsgi import app, mongo
from datetime import datetime
from jinja2 import Template, environment
import json
import os


def to_pretty_json(value):
    return json.dumps(value, sort_keys=True, indent=4, separators=(',', ': '))


environment.DEFAULT_FILTERS['pretty'] = to_pretty_json


DATE_FIELDS = [
    'started_at',
    'finished_at',
]


def json_prepare(obj):
    if obj.get('_id'):
        del obj['_id']

    for field in DATE_FIELDS:
        if obj.get(field) and type(obj[field]) == datetime:
            obj[field] = obj[field].isoformat()

    return obj


@app.route('/v1/execution/<id>/summary', methods=['GET'])
def execution_template(id):
    # load values
    collection = mongo.db[app.config['EXECUTION_COLLECTION']]

    try:
        exc = next(collection.find({'id': id}))
    except StopIteration:
        raise ModelNotFoundError(
            'Specified execution never existed, and never will'
        )

    execution = json_prepare(exc)

    if 'process_name' not in exc:
        return 'Not supported for old processes', 409

    # prepare default template
    default = []
    values = execution['values']
    for key in values:
        token = '<div><h2>{}</h2><pre>{{{{ {} | pretty }}}}</pre></div>'
        default.append(token.format(key, key))

    template = Template(''.join(default))

    # load template
    template_dir = app.config['TEMPLATE_PATH']
    process_name = execution['process_name']
    name, version, _ = process_name.split('.')

    files = os.listdir(template_dir)

    template_name = None
    for filename in files:
        try:
            fname, fversion, _ = filename.split('.')
        except ValueError:
            # Templates with malformed name, sorry
            continue

        if fname == name and fversion == version:
            template_name = filename

    if template_name:
        with open(os.path.join(template_dir, template_name), 'r') as contents:
            template = Template(contents.read())

    # return template interpolation
    return template.render(**values)



from case_conversion import pascalcase
from datetime import datetime
from functools import reduce
from operator import and_
import ast
import re
import numbers

from cacahuate.errors import InvalidInputError, RequiredListError
from cacahuate.errors import MisconfiguredProvider, RequiredIntError
from cacahuate.errors import RequiredFloatError, RequiredStrError
from cacahuate.errors import RequiredInputError, InvalidDateError
from cacahuate.xml import get_text

INPUTS = [
    'text',
    'password',
    'checkbox',
    'radio',
    'file',
    'datetime',
    'date',
    'select',
    'int',
    'float',
    'link',
]


class Option:

    def __init__(self, element):
        self.label = get_text(element)
        self.value = element.getAttribute('value')

    def to_json(self):
        return {
            'label': self.label,
            'value': self.value,
        }


class Input:
    """docstring for Input"""

    def __init__(self, element):
        self.type = element.getAttribute('type')
        self.required = element.getAttribute('required') == 'required'
        self.name = element.getAttribute('name')
        self.default = element.getAttribute('default') or None
        self.label = element.getAttribute('label') or self.name
        self.hidden = element.getAttribute('hidden') == 'true'

        self.dependencies = []

        deps_el = element.getElementsByTagName('dependencies')

        if deps_el:
            for dep_el in deps_el[0].getElementsByTagName('dep'):
                self.dependencies.append(get_text(dep_el))

    def validate(self, value, form_index):
        value = value or self.get_default()

        if self.required and (value == '' or value is None):
            raise RequiredInputError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        return value

    def get_default(self):
        return self.default

    def make_caption(self, value):
        return value

    def to_json(self):
        return {
            'type': self.type,
            'name': self.name,
            'label': self.label,
            'value': None,
            'value_caption': None,  # This is computed in node.py
            'hidden': self.hidden,
        }


class TextInput(Input):
    pass


class PasswordInput(TextInput):

    def make_caption(self, value):
        return '******'


class IntInput(Input):

    def get_default(self):
        try:
            return int(self.default)
        except ValueError:
            return 0
        except TypeError:
            return 0

    def validate(self, value, form_index):
        if self.required and type(value) != int and not value:
            raise RequiredInputError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        if type(value) != int and not value:
            value = self.get_default()

        try:
            value = int(value)
        except ValueError:
            raise RequiredIntError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        return value

    def make_caption(self, value):
        return str(value)


class FloatInput(Input):

    def get_default(self):
        if not self.default:
            return 0

        try:
            return float(self.default)
        except ValueError:
            return 0

    def validate(self, value, form_index):
        # allowed types are None, String, and Numbers
        curated = None

        if not value:
            curated = None
        elif isinstance(value, numbers.Number) or type(value) == str:
            try:
                curated = float(value)
            except ValueError:
                raise RequiredFloatError(
                    self.name,
                    'request.body.form_array.{}.{}'.format(
                        form_index,
                        self.name
                    )
                )
        else:
            raise RequiredFloatError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        if curated is None and self.required:
            raise RequiredInputError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        if curated is None:
            curated = self.get_default()

        return curated

    def make_caption(self, value):
        return str(value)


class FiniteOptionInput(Input):

    def __init__(self, element):
        super().__init__(element)

        self.options = []

        opttag = element.getElementsByTagName('options')

        if len(opttag) > 0:
            for opt in opttag[0].getElementsByTagName('option'):
                self.options.append(Option(opt))

    def __contains__(self, item):
        for opt in self.options:
            if opt.value == item:
                return True

        return False

    def make_caption(self, value):
        for opt in self.options:
            if opt.value == value:
                return opt.label


class CheckboxInput(FiniteOptionInput):

    def validate(self, value, form_index):
        super().validate(value, form_index)

        if value is None:
            value = []

        if type(value) == str:
            try:
                value = ast.literal_eval(value)
            except SyntaxError:
                value = []

        if type(value) is not list:
            raise RequiredListError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, value)
            )

        for val in value:
            if val not in self:
                raise InvalidInputError(
                    self.name,
                    'request.body.form_array.{}.{}'.format(
                        form_index,
                        self.name
                    )
                )

        return value

    def make_caption(self, value):
        def find_cap(val):
            for opt in self.options:
                if opt.value == val:
                    return opt.label

        return ', '.join(find_cap(val) for val in value)


class RadioInput(FiniteOptionInput):

    def validate(self, value, form_index):
        super().validate(value, form_index)

        curated = None

        if type(value) is str:
            curated = value

        if curated not in self:
            curated = None

        if self.required and not curated:
            raise RequiredStrError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        return curated


class SelectInput(RadioInput):
    pass


class FileInput(Input):

    def __init__(self, element):
        super().__init__(element)

        self.provider = element.getAttribute('provider')

    def validate(self, value, form_index):
        super().validate(value, form_index)

        if not value:
            value = None

        if value is not None and type(value) is not dict:
            raise InvalidInputError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        provider = self.provider

        if not value and self.required:
            raise RequiredInputError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        if provider == 'doqer':
            if value is not None:
                valid = reduce(
                    and_,
                    map(
                        lambda attr: attr in value and value[attr] is not None,
                        ['id', 'mime', 'name', 'type']
                    )
                )

                if not valid:
                    raise InvalidInputError(
                        self.name,
                        'request.body.form_array.{}.{}'.format(
                            form_index,
                            self.name
                        )
                    )
        else:
            raise MisconfiguredProvider(
                'File provider `{}` not implemented'.format(provider)
            )

        return value


class DatetimeInput(Input):

    def validate(self, value, form_index):
        super().validate(value, form_index)

        value = value or self.get_default()

        if not value:
            if self.required:
                raise RequiredInputError(
                    self.name,
                    'request.body.form_array.{}.{}'.format(
                        form_index,
                        self.name
                    )
                )

            return None

        try:
            datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")
        except TypeError:
            raise InvalidDateError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )
        except ValueError:
            raise InvalidDateError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        return value

    def get_default(self):
        if self.default == 'now':
            return datetime.now().isoformat() + 'Z'

        return None

    def make_caption(self, value):
        if not value:
            return

        return datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%fZ').strftime('%c')


class DateInput(DatetimeInput):
    pass


class LinkInput(Input):

    def validate(self, value, form_index):
        super().validate(value, form_index)

        curated = None

        if not value:
            curated = None
        elif isinstance(value, dict):
            if 'label' not in value or 'href' not in value:
                raise InvalidInputError(
                    self.name,
                    'request.body.form_array.{}.{}'.format(
                        form_index,
                        self.name
                    )
                )

            label = value['label']
            href = value['href']

            link_valid = re.match('^(https?://)[a-z0-9.]+/?$', href)

            if not link_valid:
                raise InvalidInputError(
                    self.name,
                    'request.body.form_array.{}.{}'.format(
                        form_index,
                        self.name
                    )
                )

            curated = {
                'label': label,
                'href': href,
            }
        else:
            raise InvalidInputError(
                self.name,
                'request.body.form_array.{}.{}'.format(
                    form_index,
                    self.name
                )
            )

        if not curated and self.required:
            raise RequiredInputError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        return curated


def make_input(element):
    ''' returns a build Input object given an Element object '''
    classattr = element.getAttribute('type')

    if classattr not in INPUTS:
        raise ValueError(
            'Class definition not found for input: {}'.format(classattr)
        )

    class_name = pascalcase(classattr) + 'Input'
    available_classes = __import__(__name__).inputs

    return getattr(available_classes, class_name)(element)
